using UnityEngine;
using System.Collections.Generic;
using System.Linq;

[Height(3.0f)]
[Radius(0.8f)]

[Agent]
[MovementBarrier]
//[BlocksLight]
public class Whale : Entity {	
	public Whale(Position pos, Position.Direction facing,  GameState gamestate) : base(pos, facing, gamestate) {
		 beams = new BeamDat[]{ 
			new BeamDat(this,(Position.Direction)0),
			new BeamDat(this,(Position.Direction)1),
			new BeamDat(this,(Position.Direction)2),
			new BeamDat(this,(Position.Direction)3)};
		
		drawcolorindex = (int)RenderLevel.Pal.Whale;
		movertype = Entity.MoverType.Whale;
		movesoundtype = SoundPlayer.SoundType.S_Whale;
		beamsource=true;
	}			
		
	public BeamDat[] beams;
	
	public const int MaxBeamLength=20;
	
	public void CalcBeams()
	{
		for (int i=0;i<4;i++)
		{
			Position.Direction d = (Position.Direction)i;

			BeamDat bd = beams[i];
			bd.oldvisiblelength = bd.visiblelength;
			bd.oldblocker=bd.blocker;
			bd.dir = d;
			bd.length=2;
			bd.visiblelength=2;
			bd.blocker=null;
						
			while  (gamestate.BlocksLight(bd.end, ref bd.blocker)==false)
			{				
				bd.length++;				
				if (bd.length>MaxBeamLength)
					break;
			}
			
			bd.visiblelength=2;
			while (bd.visiblelength<bd.length)
			{
				if (gamestate.At(bd.visibleend).Any( e => e.beamsource))				
					break;
				bd.visiblelength++;
			}
			
			bd.length--;
			bd.visiblelength--;
		}
	}
	
	override public bool PostMoveMessage(Entity e, Position.Direction d, MoverType othertype, float delay) 	
	{
		Position dp = d.ToPosition();
		//Position oldpos = e.pos-d;
		Position targetpos = e.pos;
		
		if (targetpos==pos)
			return false;
		if (e.pushesbeam==false)
			return false;

		if (
			(dp.x!=0 && targetpos.x==pos.x)||
			(dp.y!=0 && targetpos.y==pos.y))
		{			
			if (beams.Any( b => b.ContainsPos(targetpos) || b.end +b.dir == targetpos ) )
			{
				if (gamestate.WasOccupied(targetpos))
					return false;
				beamcollide=true;
				bool success=TryMove(d,othertype,delay);
				if (success)
				{					
					if (e.GetType()==typeof(Player))
					{
						gamestate.day=false;
					}
				}
				beamcollide=false;
				return success;
			}
		}
		
		return false;
	}
	
	private bool beamcollide=false;
	
	override public bool TryMove(Position.Direction d, MoverType othertype, float delay) 
	{
		float nextdelay = delay + TimeToCollision(othertype,beamcollide? MoverType.Beam :movertype);
		gamestate.BroadcastMessage(this,d, movertype,nextdelay);
		Position targetpos = pos + d;
		bool blockmovement = gamestate
								.At( targetpos )
								.Any( e => e.movementbarrier );
		
		bool canmove = blockmovement==false && lastmovedir==Position.Direction.MAX;
		if (canmove)
		{
			Translate(d,nextdelay);			
			gamestate.CalcBeams();
			gamestate.PostMoveMessage(this,d,movertype,nextdelay);
			return true;
		}
		return false;
	}
	
	
	public override Entity Clone(GameState gs)
	{
		return new Whale(new Position(pos), facing,gs);
	}
}
