using UnityEngine;
using System.Collections.Generic;
using System.Linq;

using System.Reflection;


abstract public class Entity : System.IComparable<Entity> {
	
	public enum MoverType {
		Keyboard=0,
		Block=1,
		Player=2,
		Wall=3,
		Whale=4,
		Beam=5,
		StaticWhale=6
	};
	
	public static float[] MoverRadius = {
		1.0f,//keyboard
		0.6f,//block
		0.5f,//player
		1.0f,//wall
		0.8f,//whale
		0.1f,//beam
	};
	
	public MoverType movertype;
	public Position pos;	
	public GameState gamestate;
	public Position.Direction facing;
	public SoundPlayer.SoundType movesoundtype;
	
	public bool beamsource=false;
	public Position.Direction lastmovedir = Position.Direction.MAX;
	public float animdelay=0;
	public Vector3 worldpos;
	public float animradius;
	public bool upper=false;
	
	public bool CanMove(Position.Direction d, int depth=0)
	{
 		Position targetpos = pos + d;
		List<Entity> atpos = gamestate.At(targetpos);
		bool blockmovement = atpos.Any(	e => e.movementbarrier );

		bool blockedbybeam;
		
		//already moved
		if (lastmovedir!=Position.Direction.MAX)
		{
		//	if (lastmovedir!=d)//why is this line here?
		//	{
				return false;
		//	}
		}
		
		if (beampushes)
		{
			//careful about modifying this ienumerable!
			var dirs = gamestate
				.GetBeamsPushing(pos)
				.Concat(gamestate.GetBeamsOverlapping(pos))
				.Select( b => b.dir);
			
			if (dirs.Contains(d.Invert()))
			{
				blockedbybeam=true;
			}
			else
			{
				blockedbybeam=false;
			}
		}
		else
		{
			blockedbybeam=false;		
		}
		
		bool canmovesister = true;//including situation with no sister
		if (facing != Position.Direction.MAX  && facing.ParallelTo(d)==false)  
		{
			if (depth==0)
			{
				Entity sister = Sister();
				if (sister==null)
				{
					Debug.LogError(GetType());
				}
				else 
				{
					canmovesister = sister.CanMove(d,depth+1);
				}
			}
		}
		
		bool canmove = blockmovement==false && blockedbybeam==false &&canmovesister;
				
		
		return canmove;
	}
		
	
	public static float TimeToCollision(MoverType a, MoverType b)
	{
		if (a==MoverType.Keyboard || a == MoverType.Beam)
			return 0;
		
		float r1 = MoverRadius[(int)a];
		float r2 = MoverRadius[(int)b];
		
		//1 = 1 unit
		float distance = 1-r1/2-r2/2;
		
		float delay = distance*Level.TurnLength;
		return delay;
	}
	
	virtual public bool TryMove(Position.Direction d, MoverType othertype, float delay) 
	{
		gamestate.BackupPositions();
		
		float nextdelay = delay + TimeToCollision(othertype,movertype);
		
		if (GetType()==typeof(Block) && othertype == MoverType.Block  && HasSister() && d==facing.Invert())
		{
			nextdelay =delay;
		}			
		
		gamestate.BroadcastMessage(this,d,movertype,nextdelay);		
		if (HasSister() && facing.ParallelTo(d)==false)//no need to propagate to sisters if pushing parallel	
		{
			gamestate.BroadcastMessage(Sister(),d,movertype,nextdelay);	
		}
		
		bool canmove = CanMove(d);
		
		if (canmove)
		{
			gamestate.DiscardPositions();
			if (HasSister() && facing.ParallelTo(d)==false)//no need to propagate to sisters if pushing parallel	
			{
				Sister().Translate(d,nextdelay);
			}
			Translate(d,nextdelay);
			gamestate.PostMoveMessage(this,d,movertype,nextdelay);				
			
			return true;
		}
		else
		{
			gamestate.RestorePositions();
			return false;
		}
	}
	
	public bool HasSister()
	{
		return facing != Position.Direction.MAX;
	}
	
	public Entity Sister()
	{
		return gamestate.entitybuckets[pos+facing].First( e => e.GetType()==GetType());
	}
	
	//be very careful about calling translate outside of this!
	public void Translate(Position.Direction d,float animdelay)
	{					
		lastmovedir = d;
		
		if (animdelay>=0)
		{
			this.animdelay=animdelay;
			if (animdelay>Level.ThisTurnLength)
			{
				Level.ThisTurnLength=animdelay;
			}
			SoundPlayer.PlaySound(movesoundtype,animdelay);
		}
		
		gamestate.RemoveEntity(this,true);				
		pos.Translate(d);
		gamestate.AddEntity(this,true);	
		
	}
	
	virtual public void PreMove() 	{}
	virtual public void PreMoveMessage(Entity e, Position.Direction d, MoverType othertype, float delay) 	{}
	virtual public bool PostMoveMessage(Entity e, Position.Direction d, MoverType othertype, float delay) 	{return false;}
	
	public bool blockslight;
	public bool agent;
	public bool movementbarrier;
	public bool beampushes;
	public bool pushesbeam;
	public int drawcolorindex;
	public float height;
	public float radius;
	
	
	private void InitParams()
	{		
		blockslight = Has<BlocksLightAttribute>();
		agent = Has<AgentAttribute>();
		movementbarrier = Has<MovementBarrierAttribute>();
		beampushes = Has<BeamPushesAttribute>();
		pushesbeam = Has<PushesBeamAttribute>();
		height = FloatParam<HeightAttribute>();
		radius = FloatParam<RadiusAttribute>();
	}
	
	private float FloatParam<T>() where T : FloatAttribute
	{
		T t = (T)System.Attribute.GetCustomAttribute(this.GetType(),typeof(T),false);
		if (t == null)
			return 0;
		return t.param;
	}	
	
	private static Dictionary<System.Type,Dictionary<System.Type,bool>> has= new Dictionary<System.Type, Dictionary<System.Type, bool>>();
	
	private bool _Has<T>()
	{
		//return Attribute.GetCustomAttribute(this.GetType(),T);
		return System.Attribute.IsDefined(this.GetType(),typeof(T),false);
	}
	
	private bool Has<T>()
	{
		System.Type thistype = this.GetType();
		
		if (has.ContainsKey(thistype)==false)
		{
			has.Add(thistype, new Dictionary<System.Type,bool>());
		}
		Dictionary<System.Type, bool> d = has[thistype];
		if (d.ContainsKey(typeof(T))==false)
		{
			d.Add(typeof(T),_Has<T>());
		}
		return d[typeof(T)];
	}
	
	protected Entity(Position pos, Position.Direction facing, GameState gamestate)
	{
		InitParams();
		this.pos = pos;
		this.facing = facing;
		this.gamestate=gamestate;
	}	
	
	//EQUALITY BEGIN
	public override bool Equals(System.Object obj)
	{
		return obj is Entity && this == (Entity)obj;
	}
	
	public override int GetHashCode()
	{
		unchecked
		{
			int hash=17;
			hash = hash*29 + pos.GetHashCode();
			hash = hash*29 + GetType().GetHashCode();
			return hash;
		}
	}
	
	public static bool operator ==(Entity a, Entity b)
	{
		if (System.Object.ReferenceEquals(a, b))
    	{
        	return true;
    	}
		
		 // If one is null, but not both, return false.
	    if (((object)a == null) || ((object)b == null))
	    {
	        return false;
	    }
		
		return 
			a.pos == b.pos &&
			a.GetType() == b.GetType();
	}
	
	public static bool operator !=(Entity x, Entity y)
	{
		return !(x==y);
	}
	//EQUALITY END
	
	//ICOMPARABLE
	public int CompareTo(Entity other)
	{
		int c1 = this.GetType().Name.CompareTo(other.GetType().Name);
		if (c1!=0)
			return c1;
		int c2 = this.pos.CompareTo(other.pos);
		return c2;
	}
	
	public override string ToString()
   	{
      	return(string.Format("({0},{1},{2})", GetType(), pos.x, pos.y));
   	}
	
	private static List<System.Type> ents = new List<System.Type>{
		typeof(Block),
		typeof(Player),
		typeof(Wall),
		typeof(Whale),
		typeof(StartTile),
		typeof(EndTile),
		typeof(StaticWhale),
	};	
	
	/// Unsigned right shift
    private static int usr(int a, int shift) {
        return (int) (((uint) a) >> shift);
    }

    private static long usr(long a, int shift) {
        return (long)(((ulong)a) >> shift);
    }		
	
	
	
	public void Write(System.IO.BinaryWriter bw)
	{	
		bw.Write((byte)ents.IndexOf(GetType()));
		bw.EncodeInt32(pos.x);
		bw.EncodeInt32(pos.y);
		bw.Write((byte)facing);
	}	
		
	public static Entity Read(System.IO.BinaryReader br, GameState gs, int version)
	{
		int typeindex = br.ReadByte();
		
		int px;
		int py;
		if (version<4)
		{
			px=br.ReadInt16();
			py=br.ReadInt16();
		}
		else
		{
			px = br.DecodeInt32();
			py = br.DecodeInt32();
		}
		
		Position.Direction facing;
		if (version>1)
		{
			facing = (Position.Direction)br.ReadByte();
		}
		else
		{
			facing = Position.Direction.MAX;
		}
		
		Position p = new Position(px,py);

		System.Type t = ents[typeindex];
		ConstructorInfo constructor = t.GetConstructor(new []{typeof(Position),typeof(Position.Direction),typeof(GameState)});
		Entity e = (Entity)constructor.Invoke(new object[]{p,facing,gs});
		return e;
	}
	
	abstract public Entity Clone(GameState gs);
}
