using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System.Linq;


public class Level : MonoBehaviour
{	
	
	public GameState gamestate = null; 
	private RenderLevel renderlevel;
	// Use this for initialization
	void Start ()
	{
		if (Application.isEditor==false)
		{			
			Screen.showCursor=false;
		}
		soundplayer = Camera.mainCamera.GetComponent<SoundPlayer>();
		GenBlankLevel();
		
	}
	
	private void SpawnWall()
	{
		for (int i=-1;i<tilewidth+1;i++)
		{
			var p1 = new Position(i,-1);
			var w1 = new Wall(p1,Position.Direction.MAX,gamestate);
			gamestate.AddEntity(w1,false);
			
			var p2 = new Position(i,tileheight);
			var w2 = new Wall(p2,Position.Direction.MAX,gamestate);
			gamestate.AddEntity(w2,false);
		}
		
		for (int i=0;i<tileheight;i++)
		{
			var p1 = new Position(-1,i);
			var w1 = new Wall(p1,Position.Direction.MAX,gamestate);
			gamestate.AddEntity(w1,false);
			
			var p2 = new Position(tilewidth,i);
			var w2 = new Wall(p2,Position.Direction.MAX,gamestate);
			gamestate.AddEntity(w2,false);
		}
	}
	
	private void GenBlankLevel()
	{
		gamestate = new GameState ();
		var player = new Player (new Position(tilewidth/2,tileheight/2),Position.Direction.MAX, gamestate);
		gamestate.AddEntity (player,false);
		
		SpawnWall();
		
		gamestate.PostMoveCalcs();
		
		TryAutomate();
		
		TakeUndoSnapshot();
		
		gamestate.SanityCheck();
	}
	
	private void GenRandomLevel()
	{
		gamestate = new GameState ();
		var player = new Player (new Position(tilewidth/2,tileheight/2),Position.Direction.MAX, gamestate);
		gamestate.AddEntity (player,false);
		
		SpawnWall();
		
		for (int i=0; i<8; i++) {
			int x = Random.Range (0, tilewidth);
			int y = Random.Range (0, tileheight);
			Position p = new Position(x,y);
			if ( gamestate.At(p).Any( e => e.movementbarrier )==false )
			{
				var w = new Wall (p, Position.Direction.MAX, gamestate);
				gamestate.AddEntity (w,false);
			}			
		}
		
		for (int i=0; i<5; i++) {
			int x = Random.Range (0, tilewidth);
			int y = Random.Range (0, tileheight);
			Position p = new Position(x,y);			
			if ( gamestate.At(p).Any( e => e.movementbarrier )==false )
			{
				var b = new Block (p, Position.Direction.MAX, gamestate);
				gamestate.AddEntity (b,false);			
			}
		}
		
		
		for (int i=0; i<5; i++) {
			int x = Random.Range (0, tilewidth);
			int y = Random.Range (0, tileheight);
			Position p = new Position(x,y);
			if ( gamestate.At(p).Any( e => e.movementbarrier )==false )
			{
				var w = new Whale (p,Position.Direction.MAX, gamestate);
				gamestate.AddEntity (w,false);			
			}
		}
		
		gamestate.PostMoveCalcs();
		
		TryAutomate();
		
		TakeUndoSnapshot();
		
		gamestate.SanityCheck();
	}
	
	static void PrintDat (string msg, GameState gs)
	{
		msg += "\n";
		foreach (Entity e in gs.entities) {
			msg += e.GetType () + "\n";
		}
		Debug.Log (msg);
	}
	
	public const float TurnLength = 0.13f;
	public static float ThisTurnLength = 0.0f;
	
	private List<byte[]> undostates = new List<byte[]>();
	
	private SoundPlayer soundplayer;
	private void DoUndo()
	{
		if (undostates.Count==0)
		{
			return;
		}
		
		byte[] buffer = undostates[undostates.Count-1];
		using (System.IO.MemoryStream ms = new System.IO.MemoryStream(buffer))
		{
			gamestate = GameState.Read(ms);
		}
		undostates.RemoveAt(undostates.Count-1);
		
		GetComponent<RenderLevel>().ResetMeshes();
		
		soundplayer.ClearSoundQueue();
//		if (Application.isEditor==false)
		TryAutomate();
	}
	
	private bool automaticrunning = false;
	public bool animating = false;	
	
	public void SetBounds()
	{	
		if (gamestate.entities.Count==0)
			return;
		int minx=gamestate.entities[0].pos.x;
		int miny=gamestate.entities[0].pos.y;
		int maxx=gamestate.entities[0].pos.x;
		int maxy=gamestate.entities[0].pos.y;
		for (int i=1;i<gamestate.entities.Count;i++)
		{
			Entity e = gamestate.entities[i];
			minx = Mathf.Min(minx,e.pos.x);
			maxx = Mathf.Max(maxx,e.pos.x);
			miny = Mathf.Min(miny,e.pos.y);
			maxy = Mathf.Max(maxy,e.pos.y);
		}
		Vector3 tl = renderlevel.PosToVect(new Position(minx,miny))-Vector3.one*Level.radius;
		Vector3 br = renderlevel.PosToVect(new Position(maxx,maxy))+Vector3.one*Level.radius;
		Vector3 center = (tl+br)/2;
		Bounds bounds = new Bounds(center,Vector3.zero);
		bounds.Encapsulate(tl);
		bounds.Encapsulate(br);
		
		Camera cam = Camera.main;
		cam.transform.position = new Vector3(bounds.center.x,-10,bounds.center.z);
		cam.orthographicSize = Mathf.Max(bounds.extents.x,bounds.extents.z/cam.aspect);
	}
	
	
	public void SetGameState(GameState gs, bool tryautomate=true)
	{
		StopCoroutines();
		automaticrunning=false;
		gamestate = gs;
				
		GetComponent<RenderLevel>().ResetMeshes();
		
		if (tryautomate)
			TryAutomate();
	}
	
	void RotateLevel()
	{
		gamestate.Rotate();
		SetGameState(gamestate);
		SetBounds();
	}
	
	void HFlipLevel()
	{
		gamestate.HFlip();
		SetGameState(gamestate);
		SetBounds();
	}
	
	void VFlipLevel()
	{
		gamestate.VFlip();
		SetGameState(gamestate);
		SetBounds();
	}
	
	bool quitting=false;
	// Update is called once per frame
	void Update ()
	{	
		
		if (Input.GetKeyDown(KeyCode.F))
		{
			Screen.fullScreen=!Screen.fullScreen;
		}
		
		
		if (quitting)
			return;
		
		if (Input.GetKeyDown(KeyCode.Escape))
		{			
			StopAllCoroutines();
			quitting=true;
			Application.LoadLevelAsync("TitleScene");
		}
			
		if (Editor.editoron)
		{
			if (Input.GetKeyDown(KeyCode.O))
			{
				RotateLevel();
			}
			if (Input.GetKeyDown(KeyCode.H))
			{
				HFlipLevel();
			}
			if (Input.GetKeyDown(KeyCode.V))
			{
				VFlipLevel();
			}
				
			if (Input.GetKeyDown(KeyCode.D))
			{
				gamestate.day = !gamestate.day;
			}
			if (Input.GetKeyDown(KeyCode.K))
			{
				gamestate.SanityCheck();
			}
			
			if (Input.GetKeyDown(KeyCode.G))
			{
				StopCoroutines();
				automaticrunning=false;
				GenRandomLevel();
			}
			
			if (Input.GetKeyDown(KeyCode.N))
			{
				StopCoroutines();
				automaticrunning=false;
				GenBlankLevel();
				SetBounds();
			}
			
		
			
		}
		
		if (gamestate.won)
			return;
		
		if (Input.GetKeyDown(KeyCode.R))
		{
			GetComponent<LoaderSaver>().Load();
			StopCoroutines();
			automaticrunning=false;
		}
		
		if (Input.GetKeyDown(KeyCode.Z))
		{
			StopCoroutines();
			automaticrunning=false;
			DoUndo();
		}
		
		
		if (automaticrunning || animating)
			return;
		
		Position.Direction movedir = Position.Direction.MAX;
		if (Input.GetButton ("Up")) {
			movedir = Position.Direction.Up;
		}
		if (Input.GetButton ("Down")) {
			movedir = Position.Direction.Down;
		}
		if (Input.GetButton ("Left")) {
			movedir = Position.Direction.Left;
		}
		if (Input.GetButton ("Right")) {
			movedir = Position.Direction.Right;
		}
		
		if (movedir != Position.Direction.MAX) {			
			TakeUndoSnapshot();
		
			if (gamestate.DoInput(movedir))
			{			
				if (gamestate.won)
				{
					TryAnimate();
					//StartCoroutine("LoadNext");
				}
				else
				{
					TryAnimate();
				}
			}
			else
			{
				DoUndo();
			}
//			GameState.Move(gamestate.GetPlayer(),movedir,gamestate,out result);
		}
	}
	
#if UNITY_EDITOR
	public const bool alloweditor=true;
#else
	public const bool alloweditor=false;
#endif
	
	public GameObject alphaplane;
	public TextMesh numbertext;
	
	
	void Awake()
	{		
		renderlevel = GetComponent<RenderLevel>();
		StartCoroutine(LoadFirstDisplay());
	}
	
	public IEnumerator LoadFirstDisplay()
	{
	//	Debug.Log("LFD");
		LoaderSaver loadersaver = GetComponent<LoaderSaver>();
		
		loadersaver.levelname = PlayerPrefs.GetString("levelname",loadersaver.levels[0]);
		int levelindex=(loadersaver.levels.ToList().IndexOf(loadersaver.levelname)+1);
		bool gameover = levelindex>=9;
		if (!gameover)
		{
			numbertext.text = levelindex.ToString() + " of 8";
		}
		else
		{
			numbertext.text = "end";
		}
		
		renderlevel.SnapColors();
		numbertext.gameObject.active=true;
		Color bc = renderlevel.GameColors[(int)RenderLevel.Pal.Beam];
		numbertext.renderer.material.color=new Color(bc.r,bc.g,bc.b,1.0f);
		
		Color c = renderlevel.GameColors[(int)RenderLevel.Pal.BG];
		alphaplane.renderer.material.color = c;
		alphaplane.active=true;
		
		iTween.Stop();
		iTween.FadeTo(numbertext.gameObject,0,1.0f);		
		iTween.ColorTo(alphaplane, new Color(c.r,c.g,c.b,0.0f),1.0f);
		yield return new WaitForSeconds(1.0f);		
		numbertext.gameObject.active=false;
		alphaplane.active=false;
	}
	
	public IEnumerator LoadNext()
	{
		LoaderSaver loadersaver = GetComponent<LoaderSaver>();
		loadersaver.levelname = PlayerPrefs.GetString("levelname",loadersaver.levels[0]);
		int levelindex=(loadersaver.levels.ToList().IndexOf(loadersaver.levelname)+2);
		bool gameover = levelindex>=9;
		if (!gameover)
		{
			numbertext.text = levelindex.ToString() + " of 8";
		}
		else
		{
			numbertext.text = "end";
		}
		numbertext.gameObject.active=true;
		Color bc = renderlevel.GameColors[(int)RenderLevel.Pal.Beam];
		numbertext.renderer.material.color=new Color(bc.r,bc.g,bc.b,0);
		iTween.FadeTo(numbertext.gameObject,1,0.5f);
		Color c = renderlevel.GameColors[(int)RenderLevel.Pal.BG];
		alphaplane.active=true;
		iTween.ColorTo(alphaplane, c,0.5f);
		yield return new WaitForSeconds(0.5f);
		if (gameover==false)
		{
			loadersaver.LoadNext();
		
			iTween.Stop();
			iTween.FadeTo(numbertext.gameObject,0,1.0f);
			iTween.ColorTo(alphaplane, new Color(c.r,c.g,c.b,0.0f),1.0f);
			yield return new WaitForSeconds(1.0f);		
			numbertext.gameObject.active=false;
			alphaplane.active=false;
		}
		else
		{
			PlayerPrefs.SetString("levelname","level13");
		}
	}
	
	public void ClearUndos()
	{
		undostates.Clear();
	}
	
	public void TakeUndoSnapshot()
	{
		using (System.IO.MemoryStream ms = new System.IO.MemoryStream())
		{
			gamestate.Write(ms);
			byte[] buffer = ms.GetBuffer().SubArray(0,(int)ms.Length);
			undostates.Add(buffer);
		}
	}
	
	public void AddEnt(Entity e) 
	{
		StopCoroutines();
		RemoveEnts(e.pos);
		gamestate.AddEntity(e,false);
		gamestate.PostMoveCalcs();
	}
	
	public void RemoveEnts(Position p)
	{
		StopCoroutines();
		List<Entity> ents = new List<Entity>(gamestate.At(p));
		if (ents.Count==0)
			return;
		
		Position.Direction removedir = Position.Direction.MAX;
		
		foreach (Entity e in ents)
		{
			gamestate.RemoveEntity(e,false);
			if (e.facing != Position.Direction.MAX)
			{
				removedir = e.facing;
				
			}
		}
		
		if (removedir!= Position.Direction.MAX)
		{
			RemoveEnts(p+removedir);
		}
		
		gamestate.PostMoveCalcs();
	}
		
	public float phasebegan;
	
	IEnumerator AnimateUpdate()
	{
		phasebegan = Time.time;
		while (Time.time - phasebegan <ThisTurnLength)
		{
			yield return 0;
		}
		
		animating=false;
		if (gamestate.won==false)
		{
			TryAutomate();
		}
		else
		{
			StartCoroutine("LoadNext");
		}
	}
	
	private void StopCoroutines()
	{
		StopAllCoroutines();
		soundplayer.ClearSoundQueue();
		animating=false;
		automaticrunning=false;
	}
	
	public void TryAnimate()
	{
		automaticrunning=false;
		animating=true;
		StartCoroutine("AnimateUpdate");			
	}
	
	public void TryAutomate()
	{				
		automaticrunning=true;
		GameState gs = gamestate.Clone();
		if (gamestate.AutomaticTick())			
		{
			TryAnimate();
		}
		else
		{	
			SetGameState(gs,false);
			automaticrunning=false;
		}
		
	}
	
	public const int tilewidth = 10;
	public const int tileheight = 10;
	public const float mapwidth = 10;
	public const float mapheight = 10;
	public const float radius = (Level.mapwidth / (2 * Level.tilewidth));// * (0.5f+(param+1.0f)/4.0f);p				
}
