using UnityEngine;
using System.Collections;
using System.Reflection;
using System.Linq;

[RequireComponent(typeof(Level))]
public class RenderLevel : MonoBehaviour {		
	
	private static Color Hex(int x, int y, int z, float a = 1.0f)
	{
		return new Color((float)x/255.0f,(float)y/255.0f,(float)z/255.0f,a);
	}
	
	private static Color Hex (int pColor)	
	{	
	    float r = ((pColor & 0xFF0000) >> 16) / 255.0f;	
	    float g = ((pColor & 0x00FF00) >> 8) / 255.0f;	
	    float b = (pColor & 0x0000FF) / 255.0f;		   	
	    return new Color(r,g,b);	
	}
	
	public enum Pal {
		BG=0,
		Wall=1,
		Block=2,
		Whale=3,
		Player=4,
		Beam=5,
		Start=6,
		End=7
	}
	
	public static Color[,] Palettes = {
		{//0
		//0 background
		Hex(220, 220, 220),
		//1 wall
		Hex(100,100,100),
		//2 block
		Hex(50, 50, 50),
		//3 whale
		Hex(100, 100, 100),
		//4 player
		Hex(160, 160, 160),
		//5 beam
		Hex(0, 0, 0,1.0f),
		//6 start
		Hex(160, 160, 160),
		//7 end
		Hex(220, 220, 220),	
		},
		{//1
		//0 background
		Hex(0x709687),
		//1 wall
		Hex(0x59270e),
		//2 block
		Hex(0x7b643b),
		//3 whale
		Hex(0x7b643b),
		//4 player
		Hex(0xe3e7e8),
		//5 beam
		Hex(0x000000),
		//6 start
		Hex(0x709687),
		//7 end
		Hex(0xe3e7e8),
		},
		{//2
		//0 background
		Hex(0x869819),
		//1 wall
		Hex(0x4b123e),
		//2 block
		Hex(0xb81f54),
		//3 whale
		Hex(0xb81f54),
		//4 player
		Hex(0xeb830c),
		//5 beam
		Hex(0xeac400),
		//6 start
		Hex(0x869819),
		//7 end
		Hex(0xeb830c),
		},
		{//3
		//0 background
		Hex(0x196467),
		//1 wall
		Hex(0x45393b),
		//2 block
		Hex(0x10929f),
		//3 whale
		Hex(0x10929f),
		//4 player
		Hex(0x0fb6c9),
		//5 beam
		Hex(0x19e1f6),
		//6 start
		Hex(0x196467),
		//7 end
		Hex(0x0fb6c9),
		},
		{//4
		//0 background
		Hex(0xff9c48),
		//1 wall
		Hex(0x006bb1),
		//2 block
		Hex(0xe60000),
		//3 whale
		Hex(0xe60000),
		//4 player
		Hex(0xf05234),
		//5 beam
		Hex(0xf9ffff),
		//6 start
		Hex(0xff9c48),
		//7 end
		Hex(0xf05234),
		},
		{//5
		//0 background
		Hex(0xe8975c),
		//1 wall
		Hex(0x3c7a7c),
		//2 block
		Hex(0xd73a46),
		//3 whale
		Hex(0xd73a46),
		//4 player
		Hex(0xde6b56),
		//5 beam
		Hex(0xe4d294),
		//6 start
		Hex(0xe8975c),
		//7 end
		Hex(0xde6b56),
		},
		{//6
		//0 background
		Hex(0xe8975c),
		//1 wall
		Hex(0x842c22),
		//2 block
		Hex(0xd73a46),
		//3 whale
		Hex(0xde6b56),
		//4 player
		Hex(0xd73a46),
		//5 beam
		Hex(0xe4d294),
		//6 start
		Hex(0xe8975c),
		//7 end
		Hex(0xd73a46),
		},
	};
	
	public Color[] GameColors =
	{
		//0 background
		Hex(220, 220, 220),
		//1 wall
		Hex(100,100,100),
		//2 block
		Hex(50, 50, 50),
		//3 whale
		Hex(100, 100, 100),
		//4 player
		Hex(160, 160, 160),
		//5 beam
		Hex(0, 0, 0,1.0f),
		//6 start
		Hex(255, 0, 0),
		//7 end
		Hex(0, 255, 0),	
	};
	
	private Mesh m;
	
	
	[HideInInspector]
	public bool failmove=false;
	
	private MeshGen maingen;
		
	
	public void ResetMeshes()
	{
		int count = level.gamestate.entities.Count + level.gamestate.beams.Count;
		//
		maingen.Alloc(4*(count+6+4),6*(count+6+4));//two extra for start/end rings, one extra for cursor
	}
	
	private Level level;
	private MeshFilter mf;
	private Camera cam;
	private Editor editor;
	void Start()
	{
		editor = GetComponent<Editor>();
		musicsource = Camera.mainCamera.gameObject.AddComponent<AudioSource>();
		musicsource.clip = music;
		musicsource.loop=true;
		musicsource.volume = 0.0f;
		musicsource.Play();
		
		cam = Camera.main;
		level = GetComponent<Level>();
		mf=GetComponent<MeshFilter>();
		m = new Mesh();
		
		m=mf.mesh;		
		maingen = new MeshGen();							
		
	}
	
	public void SnapColors()
	{
		bool day = level==null?true : level.gamestate.day;
		int targetpal = 0;
		if (day)
		{
			targetpal = 4 + ((int)(Time.time/seasonlength))%3;
		}
		
		for (int i=0;i<GameColors.Length;i++)
		{
			if (day)
			{
				Color targetcola = Palettes[targetpal,i];
				Color targetcolb = Palettes[4+((targetpal-4+1)%3),i];
				float progress = (Time.time/seasonlength)%1.0f;
				Color target = Color.Lerp(targetcola,targetcolb,progress);
				GameColors[i] = target;//Color.Lerp(GameColors[i],target,Time.fixedDeltaTime*colorlerpspeed);
			}
			else
			{
				GameColors[i] = Palettes[0,i];//,Time.fixedDeltaTime*colorlerpspeed);
			}
		}	
	}
	
	Vector3 bl = new Vector3 (-1, 0, -1) * Level.mapwidth / 2;
	Vector3 dx = new Vector3 (0, 0, Level.mapwidth / Level.tilewidth);
	Vector3 dy = new Vector3 (Level.mapheight / Level.tileheight, 0, 0); 
	Vector3 dz = new Vector3 (0, -Level.mapheight / Level.tileheight, 0)/20; 	
	
	public Vector3 PosToVect(Position p)
	{
		return bl + (p.x + 0.5f) * dx + (p.y + 0.5f) * dy;//+offset;
	}
	
	public Position VectToPos(Vector3 v)
	{
		Vector3 d = v-bl;
		float x = Vector3.Dot(d,dx);
		float y = Vector3.Dot(d,dy);
		int xi = Mathf.FloorToInt(x);
		int yi = Mathf.FloorToInt(y);
		return new Position(xi,yi);
	}
	
	float EntProgress(Entity e)
	{
		if (!level.animating)
			return 1;
		float turnprogress = Time.time-level.phasebegan;
		float myprog = (turnprogress-e.animdelay)/Level.TurnLength;
		float progress = Mathf.Clamp(myprog,0.0f,1.0f);
		return progress;
	}
	
	void CalcEntPos(Entity e)
	{
		float progress=EntProgress(e);
		Position p = e.pos-e.lastmovedir;
		Position lmd = e.lastmovedir.ToPosition();
		Vector3 diroffset = lmd.x*dx+lmd.y*dy;
		e.worldpos = bl+ (p.x+0.5f)*dx+(p.y+0.5f)*dy+diroffset*progress;
		if (e.upper)
		{
			e.worldpos=e.worldpos-dz;
		}
		float maxradius = (e.lastmovedir == Position.Direction.MAX || progress==0 )? e.radius :  e.radius + (1.0f-e.radius)/2.0f;
		e.animradius = Mathf.Lerp(maxradius,e.radius,progress);
	}
	
	
	void DrawEnts()
	{				
		//Entity pl = level.gamestate.Player();
		
		//playerpos = bl + (p.pos.x+0.5f)*dx + (p.pos.y+0.5f)*dy;
			
		
		//	float frameprogress = Mathf.Min((Time.time-level.phasebegan)/Level.TurnLength,1.0f);
		
		foreach(Entity e in level.gamestate.entities)
		{			
			Vector3 ownerpos = e.worldpos;
			
			if (e.facing == Position.Direction.MAX)
			{
				
				if (e.upper==false)
				{
					maingen.AddTile(
						ownerpos,
						Level.radius*e.animradius,
						GameColors[e.drawcolorindex]
						);						
				}
				else
				{		
					maingen.AddRing(
						ownerpos,
						Level.radius*e.animradius*0.6666666f,
						Level.radius*e.animradius,
						GameColors[e.drawcolorindex]
						);						
				}
			}
			else
			{
				maingen.AddFacingTile(
					ownerpos,
					Level.radius*e.animradius,
					Level.radius,				
					e.facing,
					GameColors[e.drawcolorindex]
					);					
			}
		}			
	}
	
	const float radiushug = 0.9f;
	const float bloat = 0.09f;
	const float worldbloat = bloat*2*Level.radius;
			
	void DrawBeams()
	{		
		//float frameprogress = Mathf.Min((Time.time-level.phasebegan)/Level.TurnLength,1.0f);
		
		foreach (BeamDat bd in level.gamestate.beams)
		{				
			
			float ownerprogress = EntProgress(bd.owner);
			Position p = bd.owner.pos-bd.owner.lastmovedir;
			Position lmd = bd.owner.lastmovedir.ToPosition();
			Vector3 diroffset = lmd.x*dx+lmd.y*dy;
			Vector3 ownerpos = bl+ (p.x+0.5f)*dx+(p.y+0.5f)*dy+diroffset*ownerprogress+dz*3;
			
			
			Position dirpos = bd.dir.ToPosition();
			Vector3 dirvect = dirpos.x*dx+dirpos.y*dy;								
			Vector3 fromOffset = dirvect/4;			
			
			//raycast
			
			//1, which tile am I centred on
			float nearestcollision=20.0f;
			
			float r1 = Level.radius*bd.owner.radius;
			

			foreach(Entity other in level.gamestate.entities)
			{				
				if (System.Object.ReferenceEquals(other,bd.owner) || (other.blockslight==false && !other.beamsource) )
					continue;
				
				Vector3 other_center3d = other.worldpos;
								
				Vector2 raycenter = new Vector2(ownerpos.z,ownerpos.x);
				Vector2 othercenter = new Vector2(other_center3d.z,other_center3d.x);
				float r2 = Level.radius*(other.radius+bloat);								
				
				float top = othercenter.y-r2;
				float bottom = othercenter.y+r2;
				float left = othercenter.x-r2;
				float right = othercenter.x+r2;
				
				float toph = othercenter.y-r2*radiushug;
				float bottomh = othercenter.y+r2*radiushug;
				float lefth = othercenter.x-r2*radiushug;
				float righth = othercenter.x+r2*radiushug;				
				
				switch (other.facing)
				{
				case Position.Direction.MAX:
				{
					break;
				}
				case Position.Direction.Up:
				{
					top=othercenter.y-Level.radius;
					toph=top;
					break;
				}
				case Position.Direction.Down:
				{
					bottom=othercenter.y+Level.radius;
					bottomh=bottom;
					break;
				}
				case Position.Direction.Left:
				{
					right=othercenter.x+Level.radius;
					righth=right;
					break;
				}
				case Position.Direction.Right:
				{
					left=othercenter.x-Level.radius;
					lefth=left;
					break;
				}
				default:
				{
					break;
				}				
				}
				
				switch(bd.dir)
				{
				case Position.Direction.Up:
				{
					//if other below,continue
					if (top>raycenter.y)
						continue;
					if (left<=raycenter.x && raycenter.x<=right)
					{
						float collisionlength = raycenter.y-bottomh;
						nearestcollision = Mathf.Min(nearestcollision,collisionlength);
					}
					break;
				}
				case Position.Direction.Down:
				{
					//if other above,continue
					if (bottom<raycenter.y)
						continue;
					if (left<=raycenter.x && raycenter.x<=right)
					{
						float collisionlength = toph-raycenter.y;
						nearestcollision = Mathf.Min(nearestcollision,collisionlength);
					}
					break;
				}
				case Position.Direction.Left:
				{
					//if other right,continue
					if (left<raycenter.x)
						continue;
					if (top<=raycenter.y && raycenter.y<=bottom)
					{
						float collisionlength = lefth-raycenter.x;
						nearestcollision = Mathf.Min(nearestcollision,collisionlength);
					}
					break;
				}
				case Position.Direction.Right:
				default:
				{
					//if other right,continue
					if (right>raycenter.x)
						continue;
					if (top<=raycenter.y && raycenter.y<=bottom)
					{
						float collisionlength = raycenter.x-righth;
						nearestcollision = Mathf.Min(nearestcollision,collisionlength);
					}
					break;
				}
				}
			}

			nearestcollision=nearestcollision-r1*0.5f+worldbloat;
			
			Vector3 start = ownerpos+fromOffset;
			Vector3 end = ownerpos+fromOffset+nearestcollision*dirvect.normalized;	
			Position rightHand = bd.dir.Rotate().ToPosition();
			Vector3 normal1 = (rightHand.x*dx+rightHand.y*dy);
			normal1 = normal1/10;
			
			Vector3 a = start+normal1;
			Vector3 b = start-normal1;
			Vector3 c = end-normal1;
			Vector3 d = end+normal1;
			
			maingen.AddQuad(a,b,c,d,GameColors[(int)Pal.Beam]);
			
		}
	}
	
	public float lerpspeed = 1.0f;
	
	void CalcEntPositions()
	{
		foreach(Entity e in level.gamestate.entities)
		{
			CalcEntPos(e);
			if (failmove && e.GetType()==typeof(Player))
			{
				e.worldpos+=new Vector3(Random.Range(-Level.radius/4,Level.radius/4),0,Random.Range(-Level.radius/4,Level.radius/4));
			}
		}
	}
	
	
	public float colorlerpspeed = 1.0f;
	public float soundlerpspeed = 1.0f;
	
	public AudioClip music;
	
	private AudioSource musicsource;
	
	private float t;
	public float seasonlength = 20.0f;
	public float targetmusicvolume = 1.0f;
	void UpdatePal()
	{
		bool day = level.gamestate.day;
		musicsource.volume = Mathf.Lerp(musicsource.volume, day ? targetmusicvolume : 0.0f, Time.fixedDeltaTime*soundlerpspeed);
		
		int targetpal = 0;
		if (day)
		{
			targetpal = 4 + ((int)(Time.time/seasonlength))%3;
		}
		
		for (int i=0;i<GameColors.Length;i++)
		{
			if (day)
			{
				Color targetcola = Palettes[targetpal,i];
				Color targetcolb = Palettes[4+((targetpal-4+1)%3),i];
				float progress = (Time.time/seasonlength)%1.0f;
				Color target = Color.Lerp(targetcola,targetcolb,progress);
				GameColors[i] = Color.Lerp(GameColors[i],target,Time.fixedDeltaTime*colorlerpspeed);
			}
			else
			{
				GameColors[i] = Color.Lerp(GameColors[i],Palettes[0,i],Time.fixedDeltaTime*colorlerpspeed);
			}
		}					
	
		
	}
	
	Vector3 Vec(Color c)
	{
		return new Vector3(c.r,c.g,c.b);
	}
	
	Color Col(Vector3 v)
	{
		return new Color(Mathf.Clamp(v.x,0,1),Mathf.Clamp(v.y,0,1),Mathf.Clamp(v.z,0,1));
	}
	
	void FixedUpdate()
	{	
		UpdatePal();
	}
	
		
	void DrawCursor()
	{
		if (Editor.editoron)
		{
			bool shiftheld = Input.GetKey(KeyCode.LeftShift) || Input.GetKey(KeyCode.RightShift);
			maingen.AddRing(
				editor.cursorvect,
				Level.radius*0.9f,
				Level.radius*1.0f,
				shiftheld ? Color.red : Color.white
				);	
		}
		else
		{
			maingen.AddRing(
				Vector3.zero,
				0,
				0,
				Color.clear);
		}		
	}
	
	//public static Vector3 playerpos;
	// Update is called once per frame
	public void LateUpdate () 
	{
		cam.backgroundColor = GameColors[(int)Pal.BG];
		
		ResetMeshes();
		CalcEntPositions();
		DrawEnts();
		
		DrawBeams();
		/*
		if (level.gamestate.day)
		{
			for (int i=0;i<maingen.vertices.Length;i++)
			{
				maingen.vertices[i] = maingen.vertices[i]+
					dx*0.05f*(Mathf.PerlinNoise(maingen.vertices[i].x+Time.time/4,maingen.vertices[i].z)-0.5f)+
					dy*0.05f*(Mathf.PerlinNoise(maingen.vertices[i].x,maingen.vertices[i].z+Time.time/4)-0.5f);
			}
		}*/
		
		DrawCursor();
		
		maingen.AssignTo(m);
	}
}
