using UnityEngine;
using System.Collections.Generic;
using System.Collections.ObjectModel;

public static class Utility {
	
	public static double Lerp(double a, double b, double t)
	{
		if (t<0) t=0;
		if (t>1) t=1;
		return (1-t)*a+t*b;
	}
	public static Position ToPosition(this Position.Direction dir)
	{
		return Position.DirPos[(int)dir];
	}
	
	public static Position.Direction ToDirection(this Position pos)
	{
		for (int i=0;i<Position.DirPos.Length;i++)
		{
			if (Position.DirPos[i].Equals(pos))
				return (Position.Direction)i;
		}
		return Position.Direction.MAX;
	}
	
	
	private static Position.Direction[] rots = {
		Position.Direction.Right,
		Position.Direction.Left,
		Position.Direction.Up,
		Position.Direction.Down,
		Position.Direction.MAX
	};
	
	
	private static Position.Direction[] hflips = {
		Position.Direction.Up,
		Position.Direction.Down,
		Position.Direction.Right,
		Position.Direction.Left,
		Position.Direction.MAX
	};
		
	private static Position.Direction[] vflips = {
		Position.Direction.Down,
		Position.Direction.Up,
		Position.Direction.Left,
		Position.Direction.Right,
		Position.Direction.MAX
	};
	
	private static Position.Direction[] invs = {
		Position.Direction.Down,
		Position.Direction.Up,
		Position.Direction.Right,
		Position.Direction.Left,
		Position.Direction.MAX
	};
	
	public static T[] SubArray<T>(this T[] data, int index, int length)
	{
	    T[] result = new T[length];
	    System.Array.Copy(data, index, result, 0, length);
	    return result;
	}
	
		
	public static Position.Direction Rotate(this Position.Direction pos)
	{
		return rots[(int)pos];
	}
	
	public static Position.Direction HFlip(this Position.Direction pos)
	{
		return hflips[(int)pos];
	}
	
	public static Position.Direction VFlip(this Position.Direction pos)
	{
		return vflips[(int)pos];
	}
	public static Position.Direction Invert(this Position.Direction pos)
	{
		return invs[(int)pos];
	}
	
	public static T RandomElement<T>(this List<T> list)
	{
		return list[Random.Range(0,list.Count)];
	}
	
		public static void Shuffle<E>(IList<E> list)
	    {
	        if (list.Count > 1)
	        {
	            for (int i = list.Count - 1; i >= 0; i--)
	            {
	                E tmp = list[i];
	                int randomIndex = Random.Range(0,i + 1);
	
	                //Swap elements
	                list[i] = list[randomIndex];
	                list[randomIndex] = tmp;
	            }
	        }
	    }
		
		public static List<E> Shuffled<E>(List<E> inputList)
		{
		     List<E> randomList = new List<E>();
		
		     int randomIndex = 0;
		     while (inputList.Count > 0)
		     {
		          randomIndex = Random.Range(0, inputList.Count); //Choose a random object in the list
		          randomList.Add(inputList[randomIndex]); //add it to the new, random list
		          inputList.RemoveAt(randomIndex); //remove to avoid duplicates
		     }
		
		     return randomList; //return the new random list
		}
		
		public static float NoteToFreq(int n)
		{
			return Mathf.Pow(2.0f, (float)n / 12.0f);
		}
	
	public static bool ParallelTo(this Position.Direction pos,Position.Direction other)
	{
		if (pos == Position.Direction.MAX || other == Position.Direction.MAX)
			return false;
		
		return pos==other || pos==other.Invert();
	}
	
	public static void AddOrdered<T>(this List<T> sortedList, T i)
	{
		if (sortedList.Count==0)
		{
			sortedList.Add(i);
			return;
		}
		int index = sortedList.BinarySearch(i);
		if (index<0)
		{
			sortedList.Insert(~index,i);
		}
	}
	
	public static void RemoveOrdered<T>(this List<T> sortedList, T i)
	{
		int index = sortedList.BinarySearch(i);
		if (index>=0)
		{
			sortedList.RemoveAt(index);
		}
	}	
	
    public static void EncodeInt32(this System.IO.BinaryWriter writer, int number)
    {
		int sign;
		byte b;
		
		sign = (number<0)?-1:0;
		
		while(true) {
			b = (byte)(number&0x7f);	
			number >>= 7;
			
			if (number == sign && (b & 0x40) == ((byte)sign & 0x40))
			{
				writer.Write(b);
				break;
			}
			b |= 0x80;
			writer.Write(b);
		}
    }

    
    public static int DecodeInt32(this System.IO.BinaryReader reader)
    {
		byte b1 = reader.ReadByte();
		byte sign = 0;
		int lvalue;
		int value;
		
		if ( (b1 & 0x80) == 0)
		{
			lvalue = b1;
			sign = (byte)(b1 & 0x40);
			if (sign!=0)
			{
				lvalue |= - (1<<7);
			}
			value=lvalue;
			return value;
		}
		byte b2 = reader.ReadByte();
		
		
		if ( (b2 & 0x80) == 0)
		{
			lvalue = b1 & 0x7f;
			lvalue |= (b2&0x7f)<<7;
			sign = (byte)(b2 & 0x40);
			if (sign!=0)
			{
				lvalue |= - (1<<14);
			}
			value=lvalue;
			return value;
		}
		
		
		byte b3 = reader.ReadByte();
		
		if ( (b3 & 0x80) == 0)
		{
			lvalue = b1 & 0x7f;
			lvalue |= (b2&0x7f)<<7;
			lvalue |= (b3&0x7f)<<14;
			sign = (byte)(b3 & 0x40);
			if (sign!=0)
			{
				lvalue |= - (1<<21);
			}
			value=lvalue;
			return value;
		}				
		
		byte b4 = reader.ReadByte();
		
		if ( (b4 & 0x80) == 0)
		{
			lvalue = b1 & 0x7f;
			lvalue |= (b2&0x7f)<<7;
			lvalue |= (b3&0x7f)<<14;
			lvalue |= (b4&0x7f)<<21;
			sign = (byte)(b4 & 0x40);
			if (sign!=0)
			{
				lvalue |= - (1<<28);
			}
			value=lvalue;
			return value;
		}
		
		
		byte b5 = reader.ReadByte();
		{
			
			lvalue = b1 & 0x7f;
			lvalue |= (b2&0x7f)<<7;
			lvalue |= (b3&0x7f)<<14;
			lvalue |= (b4&0x7f)<<21;
			lvalue |= (b5&0x7f)<<28;
			value=lvalue;
			return value;
		}
		
    }

	
//	public static List<T> ToList<T>(this ReadOnlyCollection<T> ilist)
//	{
//		return new List<T>(ilist);
//	}
}
